/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
'use strict';
import { createScanner } from '../parser/htmlScanner';
import { MarkedString } from 'vscode-languageserver-types';
import { getAllDataProviders } from '../languageFacts';
import { TokenType } from '../htmlLanguageTypes';
export function doHover(document, position, htmlDocument) {
    let offset = document.offsetAt(position);
    let node = htmlDocument.findNodeAt(offset);
    if (!node || !node.tag) {
        return null;
    }
    let dataProviders = getAllDataProviders().filter(p => p.isApplicable(document.languageId));
    function getTagHover(currTag, range, open) {
        currTag = currTag.toLowerCase();
        for (let provider of dataProviders) {
            let hover = null;
            provider.provideTags().forEach(tag => {
                if (tag.name.toLowerCase() === currTag.toLowerCase()) {
                    const tagLabel = open ? '<' + currTag + '>' : '</' + currTag + '>';
                    const tagDescription = tag.description || '';
                    hover = { contents: [{ language: 'html', value: tagLabel }, MarkedString.fromPlainText(tagDescription)], range };
                }
            });
            if (hover) {
                return hover;
            }
        }
        return null;
    }
    function getTagNameRange(tokenType, startOffset) {
        let scanner = createScanner(document.getText(), startOffset);
        let token = scanner.scan();
        while (token !== TokenType.EOS && (scanner.getTokenEnd() < offset || scanner.getTokenEnd() === offset && token !== tokenType)) {
            token = scanner.scan();
        }
        if (token === tokenType && offset <= scanner.getTokenEnd()) {
            return { start: document.positionAt(scanner.getTokenOffset()), end: document.positionAt(scanner.getTokenEnd()) };
        }
        return null;
    }
    if (node.endTagStart && offset >= node.endTagStart) {
        let tagRange = getTagNameRange(TokenType.EndTag, node.endTagStart);
        if (tagRange) {
            return getTagHover(node.tag, tagRange, false);
        }
        return null;
    }
    let tagRange = getTagNameRange(TokenType.StartTag, node.start);
    if (tagRange) {
        return getTagHover(node.tag, tagRange, true);
    }
    return null;
}
//# sourceMappingURL=htmlHover.js.map